# ! /opt/local/bin/python
# -*- coding: utf-8 -*-
import warnings

# PyNX - Python tools for Nano-structures Crystallography
#   (c) 2017-present : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr

import numpy as np
from .runner import CDIRunner, CDIRunnerException, CDIRunnerScan, params_generic
from pynx.cdi import *

params_beamline = {'size': 128, 'ndim': 3, 'obj_shape': 'icosahedron', 'obj_radius': 25,
                   'formula_phase': None, 'nb_photons': 1e8,
                   'nb_raar': 600, 'nb_hio': 0, 'nb_er': 200, 'nb_ml': 0,
                   'instrument': 'Simulation', 'scan': 1}

helptext_beamline = """
Script to perform a CDI reconstruction using a simulated dataset.
command-line/file parameters arguments: (all keywords are case-insensitive):

    size=200: size of the dataset along each dimension [default 128]

    ndim=mpx1x4: number of dimensions for the data (2 or 3, default is 3)

    obj_shape=cube: either "circle" (same a "sphere"), "square" (same as "cube"),
            "tetrahedron", "octahedron", "dodecahedron" or "icosahedron" (note that
            the latter four shapes are generated as smooth super-spheres following
            Onaka, Phil. Mag. Lett. 86 (2006), 175)

    obj_radius=56: object radius (half-size) in pixels
    
    nb_photons=1e6: number of photons in the simulated dataset [default=1e8]

    formula_phase="2 * (x*x-0.2*x*x*x + 0.5*y*y-0.1*y*y*y + 0.7*z*z)": an optional
        formula to add a phase to the object. In this formula, only sqrt,
        abs, sin, cos, tan, +, -, *, / are allowed in addition to x,y,z, which
        are coordinates relative to the object radius (i.e. varying normally from 
        -1 to +1). The computed phase is in radians.
        Another example: formula_phase="1*cos(8*(x*x+0.5*y*y))"
        [default=None, the object is real and positive]
"""

params = params_generic.copy()
for k, v in params_beamline.items():
    params[k] = v


class CDIRunnerScanSimul(CDIRunnerScan):
    def __init__(self, params, scan, timings=None):
        super(CDIRunnerScanSimul, self).__init__(params, scan, timings=timings)

    def load_data(self):
        sh = [self.params['size']] * self.params['ndim']
        pix = self.params['pixel_size_detector']
        w = self.params['wavelength']
        d = self.params['detector_distance']
        iobs = np.empty(sh)
        init_free_pixel_mask = False
        if self.cdi is None:
            init_free_pixel_mask = True
            self.cdi = CDI(iobs=iobs, pixel_size_detector=pix, wavelength=w, detector_distance=d)
        s = self.params['obj_shape']
        n = self.params['obj_radius']
        if n > sh[0] // 2:
            warnings.warn(f"CDIRunnerScanSimul: data size={sh[0]} and obj_radius={n} - there may "
                          f"be oversampling issues (2*radius > size/2)")
        fp = self.params['formula_phase']
        self.cdi = Calc2Obs() * InitObjShape(shape=s, radius=n, formula=None, formula_phase=fp) * self.cdi
        # need to shift so data is centred on the array as for any experimental data
        self.iobs = self.cdi.get_iobs(shift=True)
        # TODO: give an option for the number of photons, and do this on GPU
        self.iobs = np.random.poisson(self.params['nb_photons'] * self.iobs / self.iobs.sum())
        self.cdi.set_iobs(self.iobs, shift=True)
        if init_free_pixel_mask:
            # otherwise it is not initialised because we created self.cdi..
            self._init_free_pixel_mask(self.iobs)

    def prepare_cdi(self):
        """
        Prepare CDI object from input data.

        :return: nothing. Creates or updates self.cdi object.
        """
        super(CDIRunnerScanSimul, self).prepare_cdi()
        # Scale initial object (unnecessary if auto-correlation is used)
        if self.params['support'] != 'auto':
            self.cdi = ScaleObj(method='F', lazy=True) * self.cdi


class CDIRunnerSimul(CDIRunner):
    """
    Class to process a series of scans with a series of algorithms, given from the command-line
    """

    def __init__(self, argv, params, ptycho_runner_scan_class):
        super(CDIRunnerSimul, self).__init__(argv, params, ptycho_runner_scan_class)
        self.help_text += helptext_beamline

    def parse_arg_beamline(self, k, v):
        """
        Parse argument in a beamline-specific way. This function only parses single arguments.
        If an argument is recognized and interpreted, the corresponding value is added to self.params

        This method should be superseded in a beamline/instrument-specific child class
        Returns:
            True if the argument is interpreted, false otherwise
        """
        if k in ['scan', 'obj_shape', 'formula_phase']:
            self.params[k] = v.replace('"', '').replace("'", '')
            return True
        elif k in ['size', 'ndim', 'obj_radius']:
            self.params[k] = int(v)
            return True
        elif k in ['nb_photons']:
            self.params[k] = float(v)
            return True
        return False

    def check_params_beamline(self):
        """
        Check if self.params includes a minimal set of valid parameters, specific to a beamiline
        Returns: Nothing. Will raise an exception if necessary
        """
        pass
