# -*- coding: utf-8 -*-

# PyNX - Python tools for Nano-structures Crystallography
#   (c) 2023-present : ESRF-European Synchrotron Radiation Facility
#       authors:
#         Vincent Favre-Nicolin, favre@esrf.fr

import argparse


def make_parser(prog: str, epilog: str, default_params: dict):
    p = default_params
    parser = argparse.ArgumentParser(prog=prog,
                                     description=__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     epilog=epilog)
    grp_in = parser.add_argument_group("Input parameters")
    grp_in.add_argument(
        '--scan', action='store', type=str, default=None,
        required=False,  # TODO: required if not loading a single-scan data (e.g. cxi)
        help='scan number e.g. --scan 5 to extract a given scan number from a series.\n'
             'Alternatively a list or range of scans can be given using "scan=12,23,45"\n'
             'or scan="range(12,25)" (note the quotes, necessary when using range with ()')
    grp_in.add_argument(
        '--maxframe', '--max-frame', '--max_frame', type=int, default=p['maxframe'],
        help='maximum number of frames to import')
    grp_in.add_argument(
        '--maxsize', '--max-size', '--max_size', type=int, default=p['maxsize'],
        help='Maximum frame size to use, while keeping center of gravity of diffraction '
             'in the cut frame center.')
    grp_in.add_argument(
        '--moduloframe', '--modulo-frame', '--modulo_frame', type=int, nargs='+',
        help='Instead of taking all sequential frames, load one every N. If two numbers are\n'
             'given (--moduloframe N1 N2), N2 must be smaller than N1 and the frames loaded\n'
             'will be those with i % N1 == N2.')

    grp_exp = parser.add_argument_group("Experimental parameters")
    grp_exp.add_argument(
        '--distance', '--detectordistance', '--detector-distance', '--detector_distance',
        type=float, default=p['detectordistance'],
        help='Detector distance in meters. Mandatory unless included in the data file.')

    grp_op = parser.add_argument_group("Object and probe configuration")
    grp_op.add_argument(
        '--obj_max_pix', '--obj-max-pix', '--objmaxpix', type=int, default=p['obj_max_pix'],
        help='maximum size in pixels for the object. This is mostly to avoid mistakes when\n'
             'e.g. scan coordinates are incorrectly scaled (*1e6) and result in absurd dimensions.')
    grp_op.add_argument(
        '--obj_margin', '--obj-margin', '--objmargin', type=int, default=p['obj_margin'],
        help='margin (in pixels) around the calculated object area. this is useful when \n'
             'refining positions, to avoid getting outside the object area.')

    grp_algo = parser.add_argument_group("Algorithms")
    grp_algo.add_argument(
        '--algorithm', type=str, default=p['algorithm'],
        help=' algorithm chain used for the optimization. This chain:\n'
             '* is divided in "steps" separated by commas, e.g: ML**50,DM**100\n'
             '* is interpreted from right to left, as a mathematical operator to an\n'
             'object on the right-hand side\n'
             '* should not contain any space, unless the whole string is given between quotes ("")\n'
             '\n There are two types of commands in the chain:\n'
             '1) commands which change basic parameters or perform some analysis:\n'
             '  * probe=1 or 0: activate or deactivate the probe optimisation (by default\n'
             '    only the object is optimised)\n'
             '  * object=1 or 0: activate or deactivate the object optimisation\n'
             '  * background=1 or 0: activate or deactivate the background optimisation.\n'
             '    When set to 1, this will initialise the background to at least 1e-2 to\n'
             '    enable the background optimisation.\n'
             '  * background_smooth=3: gaussian sigma for smoothing the updated background\n'
             '    [default:3, large values are possible and will use FFT convolution]\n'
             '  * position=N or 0: activate or deactivate (0) position optimisation every N cycles\n'
             '    (preferably for AP or ML, some prior convergence is needed. Can also work\n'
             '    with DM but is not recommended). Recommended value is every 5 cycles\n'
             '  * pos_mult=1: multiplier for the calculated position shift. Can be used to accelerate\n'
             '    convergence or make it more cautious. (default:5, suitable for an object\n'
             '    with reasonable contrast)\n'
             '  * pos_max_shift: maximum shift of position update in pixels between iterations (default:2)\n'
             '  * pos_min_shift: minimum shift of position update in pixels between iterations (default:0)\n'
             '  * pos_threshold: if the integrated norm of the object gradient multiplied by the probe\n'
             '    lower than the average value (for all positions) multiplied by this threshold,\n'
             '    the position is not changed. This allows to avoid updating positions in areas\n'
             '    where the object is flat, and sensitivity to shifts is low. [default:0.2]\n'
             '  * nbprobe=3: change the number of modes for the probe (can go up or down)\n'
             '  * regularization=1e-4: setting the regularization parameter for the object, to penalize\n'
             '    local variations in ML runs and smooth the solution\n'
             '  * obj_smooth=1.5 or probe_smooth=1.0: these parameters will partially smooth the object\n'
             '    and probe, softening the resulting arrays. This applies to DM and AP algorithms.\n'
             '  * obj_inertia=0.01 or obj_probe=0.001: these parameters set the inertia of the object\n'
             '    and/or probe update, yielding more stable result. This applies to DM and AP algorithms.\n'
             '  * ortho: will perform orthogonalisation of the probe modes. The modes are sorted by\n'
             '    decreasing intensity.\n'
             '  * analysis: perform an analysis of the probe (propagation, modes). Useful combined\n'
             '    with "saveplot" to save the analysis plots\n'

             '2) operators which indicate the actual algorithm to apply to the object and probe:\n'
             '  * AP: alternate projections. Slow but converging algorithm\n'
             '  * DM: difference map. Fast early convergence, oscillating after.\n'
             '  * ML: maximum likelihood conjugate gradient (Poisson-noise). Robust, converging,\n'
             '    for final optimization.\n'
             '  These operators can be combined mathematically, e.g.:\n'
             '  * DM**100: corresponds to 100 cycles of difference map\n'
             '  * ML**40*DM**100: 100 cycles of DM followed by 40 cycles of ML (note the order)\n'
             ' \n Example algorithms chains:\n'
             '  * --algorithm ML**40,DM**100,probe=1: activate probe optimisation, then 100 DM and 40 ML\n'
             '  * --algorithm ML**100,DM**200,nbprobe=3,ML**40,DM**100,probe=1,DM**100 : first DM with\n'
             '    object update only,  then 100 DM also updating the probe, then use 3 probe modes\n'
             '    and do 100 DM followed by 40 ML\n'
             '  * --algorithm ML**100*AP**200*DM**200,probe=1: 200 DM then 200 AP then 100 ML (one step)'
             '  * --algorithm "(ML**10*DM**20)**5,probe=1":'
             '   repeat 5 times [20 cycles of DM followed by 5 cycles of ML] \n'
             '    (note the quotes necessary for the parenthesis)\n'
    )
